/* c't Mikrocontroller-im-LAN Demo Source
 * Datei: uart.c
 * Info: Routinen zur seriellen Kommunikation
 * Autor: Benjamin Benz (bbe@heise.de)
 * Datum: 24.06.04
 *
 * $LastChangedDate: 2005-12-17 15:40:32 +0100 (Sat, 17 Dec 2005) $
 * $LastChangedRevision: 24 $
 * $LastChangedBy: wb $
 * $HeadURL: svn://127.0.0.1/ct-Netzschalter/trunc/ct-netz-schalter-mcu/lib/uart.c $
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#if (__AVR_LIBC_VERSION__ < 10400UL)
#include <avr/signal.h>
#endif
#include <avr/sleep.h>
#include <avr/wdt.h>
//#include "mcu.h"
#include "uart.h"
#include "uartdefs.h"


char uart_timeout = 0;

// =================================================
/* SendeBuffer Verwaltung */
#define UART_TX_BUFFER_SIZE      128
#define UART_TX_BUFFER_MASK ( UART_TX_BUFFER_SIZE - 1 )
#if ( UART_TX_BUFFER_SIZE & UART_TX_BUFFER_MASK )
        #error TX buffer size is not a power of 2
#endif
static unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile unsigned char UART_TxHead;
static volatile unsigned char UART_TxTail;

/* EmpfangsBuffer Verwaltung */
#define UART_RX_BUFFER_SIZE     32
#define UART_RX_BUFFER_MASK ( UART_RX_BUFFER_SIZE - 1 )
#if ( UART_RX_BUFFER_SIZE & UART_RX_BUFFER_MASK )
        #error RX buffer size is not a power of 2
#endif
static unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
static unsigned char UART_RxErrs[(UART_RX_BUFFER_SIZE+1)/2];
static volatile unsigned char UART_RxHead;
static volatile unsigned char UART_RxTail;
static volatile char chars_rx = 0;

// =================================================
/* Sende-/Empfangsrichtung */
//char uart_dest_akt = DEST_XPORT; // Hier sichern wir den aktuellen Zustand
//char uart_dest_old = DEST_XPORT; // Hier sichern wir den letzten Zustand

// =========================================================================
/* Intialisierung */
void uart_init(long br, char pty, char bits, char stop)
{
        unsigned long   ubrr = (unsigned long)br;

//        DDRB  |= 0x07;  // Multiplexleitungen seriell PB0-2 als Ausgang
//        DDRB  |= 0x10;  // PORTB4 als CTS Ausgang (Y1) wenn PORTB1 auf 1, PORTB3 CTS in von WiPort(Y0) (Wenn PORTB1 auf 0)
//        uart_dest(DEST_PASSTHROUGH);
// Macht rger       PULLUPPORT |= PULLUPPIN;           // Pull up on RX line

        UCSRA = 0;  // Reset U2X
        ubrr  = ((F_CPU/16)/ubrr);   // (XTAL/(BAUDRATE*16) - 1)

        if(ubrr < 2)
        {
            UCSRA = _BV(U2X);
            ubrr = ((F_CPU + 4 * ubrr) / 8 / ubrr);
        }

        ubrr--;
        ubrr &= (4096-1);   // maximum value allowd in UBRR
        UBRRH = ((ubrr & 0xFF00) >> 8); //0x0;   // UART auf 9600 baud
        UBRRL = (ubrr & 0xFF); //0x5F;  //  UBRR = (fquarz/ (16 * BAUD) ) -1

        switch(pty)
        {
        case 'N':
        case 'n':
                pty = 0;
                break;
        case 'E':
        case 'e':
                pty = 2;
                break;
        case 'O':
        case 'o':
                pty = 3;
                break;
        default: // is No parity
                pty = 0;
        }

        bits -= 5;
        // Definiert BitPerCharacter, Parity und Anzahl Stopbits
        UCSRC = (1<<URSEL)|((bits)<<UCSZ0)|((stop-1)<<USBS)|(pty<<UPM0);

        /* Flush receive buffer */
        uart_flush_rx();

        // Flush Tx Buffer
        UART_TxTail = 0;
        UART_TxHead = 0;

        // Sender und Empfaenger freigeben, Empfangsnterrupt freigeben. TXInt wird durch send_data+Interrupt verwaltet.
        UCSRB |= ((1<<RXEN) | (1<<TXEN)|(1<<RXCIE)); //Transmit&Receive Enable + RX Int
}

/* Flush receive buffer */
void  uart_flush_rx(void)
{
        UART_RxTail = 0;
        UART_RxHead = 0;
        chars_rx = 0;
}

/* Interrupt Handler for UART Transmit Complete Interrupt */
#if defined(UNUSED)
#if (__AVR_LIBC_VERSION__ < 10400UL)
SIGNAL (SIG_UART_TRANS)
#else
ISR(USART_TXC_vect)
#endif
{
     UCSRB &= ~(1<<TXCIE);   // Interrupt sperren
}
#endif

/* Interrupt Handler for UART Data Register Empty Interrupt */
#if (__AVR_LIBC_VERSION__ < 10400UL)
SIGNAL (SIG_UART_DATA)
#else
ISR(USART_UDRE_vect)
#endif
{
    if (UART_TxHead == UART_TxTail)  // dann leer
           UCSRB &= ~(1<<UDRIE);     // disable TX int
    else
    {
           UDR   = UART_TxBuf[UART_TxTail++];    // Zeichen in Senderegister
           UART_TxTail &= UART_TX_BUFFER_MASK; // Zeiger weitersetzen.
    }
}

/* Interrupt Handler for UART Receive Complete Interrupt */
#if (__AVR_LIBC_VERSION__ < 10400UL)
SIGNAL (SIG_UART_RECV)
#else
ISR(USART_RXC_vect)
#endif
{
        unsigned char data, err;
        unsigned char tmphead;

        /* read error flag and Bit9 first, it will be cleared by reading data */
        err  = (UCSRA & ((1<<DOR)|(1<<FE)|(1<<PE)))>>1;
        // Fehlerbits sind jetzt um ein Bit nach rechts verschoben
//        err |= ((UCSRB >> RXB8) & 0x01);
        // Bit #0 von err ist jetzt das 9te Datenbit
        data = UDR;                 /* Read the received data */

        /* Calculate buffer index */
        tmphead = ( UART_RxHead + 1 ) & UART_RX_BUFFER_MASK;
        UART_RxHead = tmphead;      /* Store new index */

        if ( tmphead == UART_RxTail )   // Empfangsbuffer voll
        {
           err |= (1<<(DOR-1)); /* ERROR! Receive buffer overflow; -1 wegen rechtverschobenen Fehlerbits */
           tmphead = UART_RxHead;  // Konsistentes Verhalten bei Hardware-Overflow und Software-Overflow.
        }
        else
        {
           chars_rx++;  // ein Zeichen mehr im Buffer
        }
        if(tmphead & 0x01)   // odd address -> place error flags in upper nibble
        {
                err <<= 4; // | FE | DOR | PE  | reserviert fr das 9te bit bei 9Bit Char
                err |= UART_RxErrs[tmphead/2] & 0x0F;
        }
        else  // even address -> lower nibble
        {
                err |= UART_RxErrs[tmphead/2] & 0xF0;
        }
        UART_RxErrs[tmphead/2] = err;  /* Store receive error in buffer */
        UART_RxBuf[tmphead]    = data; /* Store received data in buffer */
}

/*
 * Ist was im Empfangbuffer?
 *     0 = nein
 *     >0  ja, anzahl der zeichen die da sind.
 *    -1 = ja, aber fehlerhaft.
 */
char uart_data_available(void)
{
    if (UART_RxHead == UART_RxTail)
       return 0;

    unsigned char tmptail = ( UART_RxTail + 1 ) & UART_RX_BUFFER_MASK;
    if(UART_RxErrs[tmptail] & ((1<<DOR)|(1<<FE)|(1<<PE)))
      return 0xff;
    return chars_rx;
}

/*
 * Hole Zeichen von serieller Schnittstelle.
 * Wartet bis Zeichen da.
 *    Returns:
 *      Positiver Wert = Zeichen wenn OK
 *      Negativer Wert = (Bit #15 gesetzt) Fehlercode im oberen Nibble, Zeichen im unteren Nibble.
 *                     Achtung: Fehlerbits um 7 nach links verschoben.
 *                     Zur Auswertung:  uart_read() & ((1<<(DOR+7))|(1<<(FE+7))|(1<<(PE+7)))
 */
int uart_read(void)
{
        unsigned char tmptail;
        union
        {
        	struct
        	{
        		unsigned int rxC     : 8;
                unsigned int dummy   : 1;
        		unsigned int pe      : 1;
        		unsigned int dor     : 1;
        		unsigned int fe      : 1;
        		unsigned int unused  : 3;
        		unsigned int errFlag : 1;
        	};
        	struct
        	{
        		unsigned char rxc;
        		unsigned char err;
        	};
        	int     value;
        } ret;

//        wdt_reset();
        while ( UART_RxHead == UART_RxTail )
        {
#if 0
             wdt_reset();
             set_sleep_mode(SLEEP_MODE_IDLE);
             sleep_mode();
#endif
             asm ("nop\n"); //wdt_reset();
             uart_timeout++;
             if(uart_timeout == 0)
             {
                uart_timeout++;
                return 0;
             }
        }
        uart_timeout = 0;
        tmptail     = ( UART_RxTail + 1 ) & UART_RX_BUFFER_MASK;/* Calculate buffer index */
        ret.err     = UART_RxErrs[tmptail/2];/* Error Flags an Bit 9 */
        ret.rxc     = UART_RxBuf[tmptail];   /* Return data */
        UART_RxTail = tmptail;               /* Store new index */
        chars_rx--;                          /* One less in buffer */
        if(tmptail & 0x01) // upper nibble if odd
        {
            ret.err >>= 4;
        }
        //err &= ((1<<DOR)|(1<<FE)|(1<<PE)|0x01);
        if(ret.err > 1)         // any of the error flags set ?
           ret.errFlag = 1;     // combined error flag
        return ret.value;       // Return data
}

/*
 * Sende Zeichen.
 * Wartet notfalls bis ein Platz im Sendbuffer frei is.
 */
void uart_send(int data)
{
    unsigned char tmphead;

    /* Calculate buffer index */
    tmphead  = ( UART_TxHead + 1 ) & UART_TX_BUFFER_MASK;
    while ( tmphead == UART_TxTail )  // TxBuf is full
    {
        UCSRB |= (1<<UDRIE);  // may have been disabled by the ISR
    }
    UART_TxBuf[UART_TxHead]     = data; /* Store data to send in buffer */
    UART_TxHead                 = tmphead;      /* Store new index */
    UCSRB |= (1<<UDRIE);  // Enable Interrupt, may have been disabled by the ISR
}

#if 0
/*
 * Schaltet die Analogschalter zwischen XPort/WiPort und RS232-Buchse um.
 */
void uart_dest(char dest)
{
        int i = 0;

        if (dest == DEST_OLD)   // Zurck zum letzten Zustand
            dest = uart_dest_old;

        // Erst warten bis Sendebuffer leer
        while(UART_TxHead != UART_TxTail)
        {
            UCSRB |= (1<<UDRIE);  // may have been disabled by the ISR
        }
        // Warten bis "USART Transmit complete"// nicht! und "USART Data Register Empty"
        // Sonst gehen die letzten Zeichen verloren
        // Der 32000-er timeout fuer das erste mal, da wird TCX nicht gesetzt :(
        while(((UCSRA & (3<<UDRE)) == 0) && (i++ < 32000))
            asm ("nop\n"); //wdt_reset();

        switch (dest)
        {
                case DEST_XPORT:
                        PORTB &= 0xFC;  // 0bxxxxxx00                   // PB0=0; PB1=0  ==> Z=Z0 Y=Y0
                        PORTB |= 0x04;  // 0b00000100                   // PB2=1 ==> (X=X1)
                        break;          // 0bxxxxx100
                case DEST_COM:
                        PORTB |=  0x03; // 0bxxxxxx11                   // PB0=1; PB1=1  ==> Z=Z1 Y=Y1
                        PORTB &=  0xFB; // 0bxxxxx0xx                   // PB2=0 ==> (X=X0) (eigentlich don't care)
                        break;          // 0bxxxxx011
                case DEST_PASSTHROUGH:
                        //PORTB |=  0x03;                       //
                        //PORTB &=  0xFC;       // 0bxxxxxx00           // PB0=1; PB1=1  ==> Z=Z0 Y=Y0
                        PORTB &=  0xF8; // 0bxxxxx000      // wb 23.12.04 => mcu TX isolated, Xport DIN&RTS <-> COM
                        break;          // 0bxxxxxx00
        }
        uart_dest_old = uart_dest_akt;
        uart_dest_akt = dest;     // Zustand sichern
}
#endif

void uart_print(char *data)
{
    if(data)
        while(*data)
            uart_send(*data++);
}

/*
 * Sende Zeichen.
 * Wartet nicht bis ein Platz im Sendbuffer frei ist.
 */
int uart_sendIfPossible(int data)
{
    unsigned char tmphead;

    /* Calculate buffer index */
    tmphead  = ( UART_TxHead + 1 ) & UART_TX_BUFFER_MASK;
    if ( tmphead == UART_TxTail )  // TxBuf is full
    {
        UCSRB |= (1<<UDRIE);  // may have been disabled by the ISR
        return 0;
    }
    UART_TxBuf[UART_TxHead]     = data; /* Store data to send in buffer */
    UART_TxHead                 = tmphead;      /* Store new index */
    UCSRB |= (1<<UDRIE);  // Enable Interrupt, may have been disabled by the ISR
    return 1;
}

int uart_printIfPossible(char *data)
{
    int len = 0;
    if(data)
        while(*data && uart_sendIfPossible(*data++))
            len++;
    return len;
}
